# hyperz.py
# hyperz catalog utility libraries
# $Log: not supported by cvs2svn $
# Revision 1.2  2007/07/17 02:31:50  tomohiro
# revise readlines to readline
#

import __builtin__
import string

import catalog
import datalist

import pylab
import math
import gaia

class Hyperz(catalog.Catalog):
    def __init__(self, file=None, zphot=None):
        """function __init__
        initialize Hyperz object
        file: Hyperz catalog file name
        zphot: zphot.param file
        """
        self.name = file
        self.column = ['ID','redshift', 'chi2', 'Pz', 'spec', 'age_rec', 'age_gyr',
          'Av', 'b', 'min99', 'max99', 'min90', 'max90', 'min68', 'max68',
          'zwm', 'Pzwm', 'Mabs', 'z2', 'Pz2']
        if zphot != None:
            config = []
            for line in __builtin__.open(zphot):
                if line[0] == '#' or line[0] == '\n': continue
                line = line.strip()
                data = line.split()
                key = data[0].strip()
                value = datalist.str2data(data[1].strip())
                config.append((key,value))
            self.zphot = dict(config)
    
    def readline(self, nv=None):
        photlist = []
        #print 'open '+self.name
        for line in __builtin__.open(self.name):
            if len(line) <= 0: break
            if line[0] == '#': continue
            #print line
            line = line.strip()
            data = line.split()
            phot = []
            for key,value in zip(self.column, data):
                value = datalist.str2data(value, nv=nv)
                phot.append((key,value))
            yield dict(phot)
#            phot = dict(phot)
#            photlist.append(phot)
#        return datalist.DataList(photlist)

    def zhist(self, col='b-o', title='photo z count', bin=0.1):
        pylab.figure()
        photlist = self.readlines()
        zlist = []
        for phot in photlist:
            z = float(phot['redshift'])
            zlist.append(z)
        zlist = pylab.array(zlist)
        max = math.ceil(zlist.max())
        min = math.floor(zlist.min())
        
        n, bins = pylab.hist(zlist, bins=pylab.arange(int(min-1), int(max+1), bin))[:2]

        pylab.title(title)
        pylab.xlabel('photo z')
        pylab.ylabel('count')
    def typehist(self, col='b-o', title='best-fit spectral type count'):
        pylab.figure()
        photlist = self.readlines()
        typelist = []
        for phot in photlist:
            type = float(phot['spec'])
            typelist.append(type)
        typelist = pylab.array(typelist)
        max = math.ceil(typelist.max())
        min = math.floor(typelist.min())
        
        n, bins = pylab.hist(typelist, bins=pylab.arange(int(min-1), int(max+1), 1))[:2]

        pylab.title(title)
        pylab.xlabel('spectral type')
        pylab.ylabel('count')

    def specplot_all(self, magfile, fltnames):
        """
        magfile: gaia file name of magnitude data
        fltnames: fltnames in magfile
        """
        ## make filter list
        fltlist = get_filterlist(self.zphot['FILTERS_RES'], self.zphot['FILTERS_FILE'])
        
        photlist = self.readlines()
        mag_f = gaia.open(magfile)
        maglist = mag_f.readlines()
        for phot, mag in zip(photlist,maglist):
            id = phot['ID']
            specplot(str(id)+".spe", fltlist, mag, phot, fltnames, str(id)+"_spe.png")
    def specplot(self, id, magfile, fltnames, outfile=None):
        """
        id: source ID
        magfile: gaia file name of magnitude data
        fltnames: fltnames in magfile
        outfile: output file name
        """
        ## make filter list
        fltlist = get_filterlist(self.zphot['FILTERS_RES'], self.zphot['FILTERS_FILE'])
        
        photlist = self.readlines()
        mag_f = gaia.open(magfile)
        maglist = mag_f.readlines()
        phot = photlist.getbykey('ID', id)
        mag = maglist.getbykey('NUMBER', id)
        specplot(str(id)+".spe", fltlist, mag, phot, fltnames, outfile)

    def catmerge(self, gcat, outcat, hzkeys=None):
        """
        merge catalog with originga gaia file
        gcat: original gaia file
        outcat: output catalog file name
        hzkeys: key name list to write (default: all)
        """
        hzlist = self.readlines()
        g_f = gaia.open(gcat)
        glist = g_f.readlines()
        if hzkeys == None:
            hzkeys = self.column
        outlist = []
        for hzdata, gdata in zip(hzlist, glist):
            outdata = []
            for gkey in gdata.keys():
                outdata.append((gkey, gdata[gkey]))
            for hzkey in hzkeys:
                outdata.append((hzkey, hzdata[hzkey]))
            outlist.append(dict(outdata))
        wkeys = g_f.column
        wkeys.extend(hzkeys)
        gaia.writedata(outlist, outcat, wkeys)

    def probplot(self, **kwargs):
        """
        plot z probability graph
        """
        logname = self.zphot['OUTPUT_FILE'] + ".log_phot"
        id = 1
        znum = (self.zphot['Z_MAX'] - self.zphot['Z_MIN'])/self.zphot['Z_STEP']
        zindex = znum
        zdata = []
        pdata = []
        for line in __builtin__.open(logname):
            line = line.strip()
            data = line.split()
            zdata.append(float(data[1]))
            pdata.append(float(data[3]))
            if zindex > 0:
                zindex = zindex - 1
            else:
                pylab.ioff()
                pylab.clf()
                pylab.plot(zdata, pdata, **kwargs)
                pylab.title('phot-z probability for ' + str(data[0]))
                pylab.savefig(str(data[0]) + "_prob.png", papertype='a10')
                zindex = znum
                zdata = []
                pdata = []
        
def open(file=None, zphot=None):
    """ open hyperz catalog file and returns Hyperz object
    file: hyperz catalog file name
    """
    return Hyperz(file, zphot)

def specplot(specfile, fltlist, magdata, zphotdata, fltnames, outfile=None):
    """bestfit plot
    specfile: *.spe file from hyperz
    fltlist: list of hyperz.Filter
    magdata: magnitude datalist of object
    zphotdata: photoz datalist of object
    fltnames: filter names in magdata
    outfile: output graphic file name
    """
    pylab.ioff()
    pylab.clf()
    wl = []
    mag = []
    ## plot spectrum
    for line in __builtin__.open(specfile):
        line = line.strip()
        data = line.split()
        wl.append(float(data[0]))
        mag.append(float(data[1]))
    wl = pylab.array(wl)
    mag = pylab.array(mag)
    #pylab.plot(wl, mag)
    #pylab.semilogx(wl[pylab.where(mag < 90)], mag[pylab.where(mag < 90)])
    pylab.plot(wl[pylab.where(mag < 90)], mag[pylab.where(mag < 90)])
    for flt,fltdata in zip(fltnames,fltlist):
        fltmag = magdata[flt]
        fltmagerr = magdata[flt+'_ERR']
        fltwl = fltdata.wl[pylab.where(fltdata.res == fltdata.res.max())][0]
        fltwlerr = fltdata.fwhm()/2.
        if (fltmag > 90 or fltmagerr > 90) : continue
        pylab.errorbar(fltwl, fltmag, fltmagerr, fltwlerr)
    (xmin, xmax, ymin, ymax) = pylab.axis()
    pylab.axis([xmin, xmax, ymax, ymin])
    pylab.title('best fit SED for '+ str(zphotdata['ID'])+ '\n' +
                ' z = '+ str(zphotdata['redshift'])+
                ' type = ' + str(zphotdata['spec'])+
                ' age = ' + str(zphotdata['age_gyr']) + 'Gyr' +
                ' Av = ' + str(zphotdata['Av']))
    pylab.xlabel('wavelength [A]')
    pylab.ylabel('mag_AB')
    if outfile != None:
        pylab.savefig(outfile)
    else:
        pylab.ion()
        pylab.draw()

class Filter:
    def __init__(self, name, wl, res):
        """
        name: filter name
        wl: wavelength (A)
        res: filter response
        """
        self.name = name
        self.wl = pylab.array(wl)
        self.res = pylab.array(res)
    def plot(self):
        pylab.plot(self.wl, self.res)
        pylab.title(self.name)
        pylab.xlabel('wavelength [A]')
        pylab.ylabel('response [%]')
    def fwhm(self):
        max = self.res.max()
        max_index = pylab.where(self.res == max)[0][0]
        for wl, res in zip(self.wl[0:], self.res[0:]):
            if res > max/2:
                wl1 = wl
                break
        for wl, res in zip(self.wl[max_index:], self.res[max_index:]):
            if res < max/2:
                wl2 = wl
                break
        return wl2 - wl1

def load_filters(file):
    file_f = __builtin__.open(file)
    fltlist = []
    while(True):
        line = file_f.readline()
        if line == "": break
        line = line.strip()
        data = line.split()
        num = int(data[0])
        name = string.join(data[1:])
        wl = []
        res = []
        for num in range(num):
            line = file_f.readline()
            line = line.strip()
            data = line.split()
            wl.append(float(data[1]))
            res.append(float(data[2]))
        #fltlist.append({'name':name, 'wl': wl, 'res': res})
        fltlist.append(Filter(name, wl, res))
    return fltlist

                       
def get_filterlist(fltresfile, fltfile):
    fltres = load_filters(fltresfile)
    outlist = []
    for line in __builtin__.open(fltfile):
        if line[0] == '#' or line[0] == '\n': continue
        line = line.strip()
        fltnum = line.split()[0]
        #print "num: "+fltnum
        outlist.append(fltres[int(fltnum)-1])
    return outlist

    
