"""
cosmology.py - tools for cosmological distances etc.
2007. 9. 12 YOSHIKAWA, Tomohiro
reference:
  Hogg, D. W. 2000, astro-ph/9905116

$Log: not supported by cvs2svn $
Revision 1.3  2008/03/19 10:24:43  tomohiro
add da, vc, tl method to Cosmology

Revision 1.2  2008/02/01 23:45:34  tomohiro
modify dc() to accept z array

Revision 1.1  2007/09/24 09:07:51  tomohiro
astronomical tool library
"""

import sys
import numpy

class Cosmology:
    def __init__(self, h=0.7, m=0.3, l=0.7, dz=0.001, maxz=10):
        """
        initialize cosmological parameters
        """
        self.h = h # Hubble constant H = 100h (km/s/Mpc)
        self.m = m # Omega matter
        self.l = l # Omega lambda
        self.dz = dz # delta z for integrate
        self.dh = 9.26e25/self.h # Hubble distance (m)
        self.th = 9.78e9/self.h # Hubble time (yr)
        self.maxz = maxz

        self.zz = numpy.arange(0, self.maxz+self.dz, self.dz)
        self.ezi = 1/numpy.sqrt(self.m*(1+self.zz)**3+self.l) # inverse of E(z)
    def dc(self, z=0):
        """
        comoving transverse (m)
        z: redshift
        """
        try:
            len(z)
            rarray = True
        except TypeError:
            z = [z]
            rarray = False
        rdc = []
        for zz in z:
            ind = numpy.arange(int(zz/self.dz)+1)
            rdc.append(self.dh*self.dz*sum((self.ezi[ind[1:]]+self.ezi[ind[:-1]])/2.))
        if rarray:
            return numpy.array(rdc)
        else:
            return rdc[0]
        #return dh*self.dz*sum(1/numpy.sqrt(self.m*(1+zz)**3+self.l))

    def dl(self, z=0):
        """
        luminosity distance (m)
        z: redshift
        """
        return (1+z) * self.dc(z)

    def da(self, z=0):
        """
        angular diameter distance (m)
        z: redshift
        """
        return self.dc(z) / (1+z)

    def vc(self, z1=0, z2=1):
        """
        comoving volume (m3/sold angle)
        z1, z2: redshift between which calculate the volume
        """

        ind = numpy.arange(int(z1/self.dz), int(z2/self.dz)+1)
        #print self.zz[ind]
        dm = self.dc(self.zz[ind])
        dm2ez = dm**2*self.ezi[ind]

        ind1 = range(len(ind))
        return self.dh*self.dz*sum((dm2ez[ind1[1:]]+dm2ez[ind1[:-1]])/2.)
        
    def tl(self, z):
        """
        look back time (yr)
        z: redshift
        """
        try:
            len(z)
            rarray = True
        except TypeError:
            z = [z]
            rarray = False
        rtl = []
        
        for zz in z:
            ind = numpy.arange(int(zz/self.dz)+1)
            rtl.append(self.th*self.dz*sum((self.ezi[ind[1:]]/(1+self.zz[ind[1:]])+self.ezi[ind[:-1]]/(1+self.zz[ind[:-1]]))/2.))
        if rarray:
            return numpy.array(rtl)
        else:
            return rtl[0]
        
def usage():
    print __doc__
    print """
usage:
   cosmology.py z1 [z2]
parameters:
   z1: redshift
   z2: calculate comving volume between z1 and z2
"""
if __name__ == "__main__":
    try:
        i = 1
        z1 = float(sys.argv[i]); i+=1
        if len(sys.argv)>2:
            z2 = float(sys.argv[i]); i+=1
        else:
            z2 = None
    except IndexError:
        usage()
        sys.exit(-1)

    csm = Cosmology()
    
    print "comoving transverse to z=%.3f       : %e (m)"%(z1, csm.dc(z1))
    print "luminosity distance to z=%.3f       : %e (m)"%(z1, csm.dl(z1))
    print "angular diameter distance to z=%.3f : %e (m)"%(z1, csm.da(z1))
    print "look back time to z=%.3f            : %e (yr)"%(z1, csm.tl(z1))

    if z2:
        print ""
        print "comoving volume between z=%.3f and z%.3f: %e (m^3/solid angle)"%(z1, z2, csm.vc(z1, z2))
