# instrument.py
# class for intrument configuration
import os.path
from mcsmdp.shape import Shape
from mcsmdp.error import *
from astrocat import configfile, gaia
import numpy

class Instrument:
    center_x=None
    center_y=None
    fov = None
    slit_area = None
    pixelscale = None
    magnify = None
    dispaxis = 1
    disperser = None
    def __init__(self, fov, slit_area, pixelscale, magnify, dispaxis, disperser):
        self.fov = fov
        self.slit_area = slit_area
        self.pixelscale = pixelscale
        self.magnify = magnify
        self.dispaxis = dispaxis
        self.center_x = fov.x
        self.center_y = fov.y
        self.disperser = disperser
    def update_center(self, x, y, ra, dec, angle):
        self.fov.update(ra,dec, angle=angle, coord='fk5', coord_unit='degree')
        self.slit_area.update(ra,dec, angle=angle, coord='fk5', coord_unit='degree')
        self.center_x = x
        self.center_y = y
    def getreg(self, name=False, slitarea=False):
        if slitarea:
            return self.fov.getreg(name=name) + ';' + self.slit_area.getreg()
        else:
            return self.fov.getreg(name=name)
            
    def getspectrum(self, slit):
        pass

    #def show(self):
    #    self.fov.show()
    #    self.slit_area.show()
    #def hide(self):
    #    self.fov.hide()
    #    self.slit_area.hide()
    #def updatebyds9(self):
    #    xpa = ds9.open()
    #    splist1 = xpa.getinfo(coord='fk5')
    #    if len(splist1) != 1:
    #        raise MCSMDPError, 'ambiguous region selection'
    #    sp1 = splist1[0]
    #    ra = sp1.x
    #    dec = sp1.y
    #    splist2 = xpa.getinfo(coord='image')
    #    sp2 = splist2[0]
    #    x = sp2.x
    #    y = sp2.y
    #    self.update_center(x, y, ra, dec, sp1.angle)

class FOV(Shape):
    """FOV: instrument field of view
    x, y: coordinate (x, y or RA, DEC)
    size_x, size_y: size of fov (x, y or RA, DEC)
    coord: coodinate name (fk5, image...)
    name: name to identify the FOV
    """
    def __init__(self, x=1792, y=1024, size_x=7, size_y=4, coord='fk5',
                 size_unit='arcmin', coord_unit='image', pa=0,
                 name='default', fov_type='fov', type='box', **kwargs):
        if coord=='fk5':
            angle = pa + 90
        else:
            angle = pa
        id = fov_type+':'+name
        self.name=name
        Shape.__init__(self, x, y, coord=coord, coord_unit=coord_unit,
                       type=type, size_x=size_x, size_y=size_y, size_unit=size_unit,
                       angle=angle, id=id, **kwargs)

    def getreg(self, rotate=1, move=1, edit=0, delete=1, name=False, name_font='courier 80 bold', **kwargs):
        regstr = ""
        if name:
            regstr = '%s;text(%s, %s) # textangle=%s text={%s} font={%s} color=%s'%(
                self.coord, str(self.x), str(self.y), str(self.angle), self.name, name_font, self.regprops['color'])
            #print regstr
            if self.id != None:
                regstr += " tag={id:%s}"%self.id
            regstr += ';'
        regstr += Shape.getreg(self, rotate=rotate, move=move, edit=edit, delete=delete)
        return regstr
   
    def update(self, x, y, angle, name=None, **kwargs):
        if name != None: self.name=name
        Shape.update(self, x=x, y=y, angle=angle, **kwargs)

    #def show(self, rotate=1, move=1, edit=0, delete=1, name=False, name_font='courier 20 bold', **kwargs):
    #    Shape.update(self, **kwargs)
    #    if name:
    #        xpa = ds9.open()
    #        regstr = '%s;text(%s, %s) # textangle=%s text={%s} font={%s} color=%s'%(
    #            self.coord, str(self.x), str(self.y), str(self.angle), self.name, name_font, self.regprops['color'])
    #        #print regstr
    #        if self.id != None:
    #            regstr = regstr+" tag={%s}"%self.id
    #        xpa.xpaset('regions', data=regstr)
    #    Shape.show(self, rotate=rotate, move=move, edit=edit, delete=delete)

class disperser:
    grism=None
    filter=None
    grism_name=None
    filter_name=None
    def __init__(self, grism, filter):
        """constructor
        grism = grism datalist
        filter = filter datalist
        """
        self.grism=grism
        self.filter=filter
        self.filter_name=['DET']

    def setdisperser(self, grism_name, filter_name):
        """ set the name of disperser elements
        grism_name = the name of disperser in the disperser list
        filter_name = a LIST of the filter names in the filter list
        """
        if self.grism.has_data('NAME', grism_name):
            self.grism_name = grism_name
        else:
            raise MCSMDPError, "no such grism name: "+grism_name

        self.filter_name=['DET']
        for flt in filter_name:
            if self.filter.has_data('NAME', flt):
                self.filter_name.append(flt)
            else:
                raise MCSMDPError, "no such filter name: "+flt

    def getfilterparam(self, filter_name=None):
        """
        return cuton and cutoff wavelength
        """
        if filter_name == None:
            filter_name = self.filter_name
        else:
            filter_name += ['DET']
        lmin_list=[]
        lmax_list = []
        for fname in filter_name:
            fdata = self.filter.getbykey('NAME', fname)
            lmin_list.append(fdata['CUT_ON'])
            lmax_list.append(fdata['CUT_OFF'])
            #print fdata['NAME'],fdata['CUT_ON'],fdata['CUT_OFF']
        lmin_list = numpy.array(lmin_list)
        lambda_min = lmin_list.max()
        lmax_list = numpy.array(lmax_list)
        lambda_max = lmax_list.min()
        return lambda_max, lambda_min
    
    def getgrismparam(self, grism_name=None):
        """
        return center wavelength, and pixel per wavelength rate
        """
        if grism_name == None:
            gdata = self.grism.getbykey('NAME', self.grism_name)
        else:
            gdata = self.grism.getbykey('NAME', grism_name)
        return  gdata['CENTER'], gdata['RATE']

    def getdispersion(self):
        """ get max/min dispersion pixel from slit position """
        if self.grism_name == None or len(self.filter_name) == 1:
            raise MCSMDPError, "grism or filter name is not yet defined"

        lambda_max, lambda_min = self.getfilterparam()
        center, rate = self.getgrismparam()
        #print gdata['NAME'],gdata['CENTER'],gdata['RATE']
        #print lambda_max,lambda_min
        spec_max = (lambda_max - center)/rate
        spec_min = (lambda_min - center)/rate
        return spec_max, spec_min

class spectrum(Shape):
    def __init__(xmax, xmin, ymax, ymin, slit_x, slit_y, slit_id, **kwargs):
        pass

def LoadFile(file):
    conf = configfile.load(file)
    x = conf['FOV_CENTER_X']
    y = conf['FOV_CENTER_Y']
    fov = FOV(x, y, size_x=conf['FOV_SIZE_X'], size_y=conf['FOV_SIZE_Y'], coord='image',
              size_unit=conf['FOV_SIZE_UNIT'], coord_unit = conf['FOV_COORD_UNIT'],
              pa=conf['FOV_PA'], name='MOIRCS', fov_type='fov',
              type=conf['FOV_SHAPE'], color=conf['FOV_COLOR'])
    slit_area = FOV(x, y, size_x=conf['SLIT_AREA_SIZE_X'], size_y=conf['SLIT_AREA_SIZE_Y'],
                    coord='image', coord_unit = conf['FOV_COORD_UNIT'],
                    size_unit=conf['SLIT_AREA_SIZE_UNIT'],
                    pa=conf['FOV_PA'], name='MOIRCS', fov_type='slit_area',
                    type=conf['SLIT_AREA_SHAPE'], color=conf['SLIT_AREA_COLOR'])
    confdir = os.path.dirname(file)
    filter_f = gaia.open(confdir+'/'+conf['FILTER_CONF'])
    grism_f  = gaia.open(confdir+'/'+conf['GRISM_CONF'])
    disp = disperser(grism_f.readlines(), filter_f.readlines())
    if conf.has_key('DEF_FILTER') and conf.has_key('DEF_GRISM'):
        disp.setdisperser(conf['DEF_GRISM'], [conf['DEF_FILTER']])
    inst = Instrument(fov, slit_area, pixelscale=conf['PIXEL_SCALE'],
                      magnify=conf['MAGNIFY'], dispaxis=conf['DISPAXIS'],
                      disperser=disp)
    return inst
