# shape.py
# classes for shape of ds9 region
# 2006.06.02 YOSHIKAWA, Tomohiro
import RO.DS9
from error import *
from astrocat import datalist

def getcu(coord_unit):
    if coord_unit == 'degree':
        return 'd'
    elif coord_unit == 'radian':
        return 'r'
    elif coord_unit == 'physical':
        return 'p'
    elif coord_unit == 'image':
        return 'i'
    else:
        return ''

def getsu(size_unit):
    if size_unit == 'arcsec':
        return '"'
    elif size_unit == 'arcmin':
        return '\''
    elif size_unit == 'degree':
        return 'd'
    elif size_unit == 'radian':
        return 'r'
    elif size_unit == 'physical':
        return 'p'
    elif size_unit == 'image':
        return 'i'
    else:
        return ''

def bool2int(bool):
    if bool: return 1
    else: return 0

class Shape:
    """Class shape
    """
    def __init__(self, x, y, coord='image', coord_unit='default',
                 type='circle', size_x=100, size_y=100, size_unit='default',
                 angle=0, point='circle',
                 length='10', vector=True,
                 x2=None, y2=None, arrow1=None, arrow2=None,
                 id=None, groupid=None, **kwargs):
        """constructor
        x,y: X, Y coordinate
        coord: coordinate name for X, Y (image, fk5, ...)
        coord_unit: (default, degree, radian, physical, image)
        type: box, ellipse, circle, point
        size_x: size for x-axial direction (box, ellipse) or radius (circle)
        size_y: size for y-axial direction (box, ellipse)
        size_unit: unit of size (default, arcsec, arcmin, degree, radian, physical, image)
        angle: for box, ellipse
        point: point type (only for point)
        id: shape identifier
        **kargs: other properties for ds9 region (color etc.)
        """
        if id != None:
            if id.find(' ') >= 0:
                raise MCSMDPError, 'invalid shape identifier: %s'%id
        self.x = datalist.str2data(str(x))
        self.y = datalist.str2data(str(y))
        self.type = type
        self.size_x = datalist.str2data(str(size_x))
        self.size_y = datalist.str2data(str(size_y))
        self.angle = angle
        self.point = point
        self.coord = coord
        self.regprops = kwargs
        # set size of coordinate (cu)
        self.cu = getcu(coord_unit)
        self.coord_unit = coord_unit
        # set size of unit (su)
        self.su = getsu(size_unit)
        self.size_unit = size_unit
        # for vector
        self.length = length
        self.vector = vector
        # for line
        self.x2 = x2
        self.y2 = y2
        self.arrow1 = arrow1
        self.arrow2 = arrow2
        # id
        self.id=id
        self.groupid=groupid
        # xpa
        self.xpa = None
        # enable
        self.enable = True

    def update(self, x=None, y=None, coord=None, coord_unit=None,
               type=None, size_x=None, size_y=None, size_unit=None,
               angle=None, point=None, length=None, vector=None,
               x2=None, y2=None, arrow1=None, arrow2=None,
               id=None, **kwargs):
        if x != None: self.x = x
        if y != None: self.y = y
        if coord != None: self.coord = coord
        if coord_unit != None:
            self.coord_unit = coord_unit
            self.cu = getcu(coord_unit)
        if type  != None: self.type = type
        if size_x != None: self.size_x = size_x
        if size_y != None: self.size_y = size_y
        if size_unit != None:
            self.size_unit = size_unit
            self.su = getsu(size_unit)
        if angle != None: self.angle = angle
        if point != None: self.point = point
        if length != None: self.length = length
        if vector != None: self.vector = vector
        if x2 != None: self.x2 = x2
        if y2 != None: self.y2 = y2
        if arrow1 != None: self.arrow1 = arrow1
        if arrow2 != None: self.arrow2 = arrow2
        
        if id != None: self.id = id
        
        for k,v in kwargs.iteritems():
            self.regprops[k] ="%s"%v
        
    def getreg(self, move=0, rotate=0, edit=0, delete=1):
        props = ""
        editprop = " move=%d rotate=%d edit=%d delete=%d"%(move,rotate,edit,delete)
        for k,v in self.regprops.iteritems():
            props += k+"="+str(v)+" "
        if(self.type == 'box' or self.type == 'ellipse'):
            regstr = "%s;%s(%s%s,%s%s,%s%s,%s%s,%s) # %s" % (
                self.coord, self.type, self.x, self.cu, self.y, self.cu,
                self.size_x, self.su, self.size_y, self.su,
                self.angle, props)
        elif(self.type == 'circle'):
            regstr = "%s;circle(%s%s,%s%s,%s%s) # %s" % (
                self.coord, self.x, self.cu, self.y, self.cu,
                self.size_x, self.su, props)
        elif(self.type == 'point'):
            regstr = "%s;point(%s%s,%s%s) # point=%s %s" % (
                self.coord, self.x, self.cu, self.y, self.cu, self.point, props)
        elif(self.type == 'vector'):
            props += "vector=%d "%(bool2int(self.vector))
            regstr = "%s;vector(%s%s,%s%s,%s%s,%s) # %s" % (
                self.coord, self.x, self.cu, self.y, self.cu,
                self.length, self.su, self.angle, props)
        elif(self.type == 'line'):
            props += "line=%d %d "%(bool2int(self.arrow1),bool2int(self.arrow2))
            regstr = "%s;line(%s%s,%s%s,%s%s,%s%s) # %s" % (
                self.coord, self.x, self.cu, self.y, self.cu,
                self.x2, self.cu, self.y2, self.cu, props)

        # set id tag
        if self.id != None:
            regstr = regstr+"tag={id:%s}"%self.id
        if self.groupid != None:
            regstr = regstr+" tag={gid:%s}"%self.groupid
        return regstr + editprop

class ShapeList:
    """class ShapeList
    base class for list of shapes
    """
    def __init__(self):
        self.coord='image'
        self.cu='i'
        self.type='circle'
        self.size_x=100
        self.size_y=100
        self.su='i'
        self.point='circle'
        self.list=[]
        self.xpa = None
        self.regmax = 500
        self.groupid=None
    def _loadcoord(self, datalist, coord='image', xname=None, yname=None):
        """method _loadcoord
        private method for loading coordinate data
        datalist: a list of lists of dictionaries
        coord: input coordinate (xy|radec)
        xname: keyword for X coordinate
        yname: keyword for Y coordinate
        """
        if xname == None or yname == None :
            if coord == 'image':
                xname='X'
                yname='Y'
            else:
                xname='RA'
                yname='DEC'
        self.coord=coord
        self.list = []
        for data in datalist:
            x=data[xname]
            y=data[yname]
            yield (x, y, xname, yname)
    
    def loaddata(self, datalist, coord='image', coord_unit='default', xname=None, yname=None, group='', **kwargs):
        """method loaddata
        datalist: a list of lists of dictionaries
        coord: input coordinate (xy|radec)
        coord_unit: (default, degree, radian, physical, image)
        xname: keyword for X coordinate
        yname: keyword for Y coordinate
        **kwargs: other properties for ds9 region
        """
        self.coord=coord
        self.cu=getcu(coord_unit)
        self.groupid=group
        index=0
        for (x, y, xname, yname) in self._loadcoord(datalist, coord, xname, yname):
            idstring = "%s%04i"%(group, index)
            index=index+1
            obj = Shape(x, y, type=self.type, size_x=self.size_x, size_y=self.size_y,
                        coord=coord, coord_unit=coord_unit, id=idstring, groupid=group, **kwargs)
            self.list.append(obj)

    def setprop(self, type=None, size_x=None, size_y=None, size_unit=None, point=None, coord=None, coord_unit=None, **kwargs):
        """method setprop
        set properties for shapes
        type: box, ellipse, circle, point
        size_x: size for x-axial direction (box, ellipse) or radius (circle)
        size_y: size for y-axial direction (box, ellipse)
        size_unit: unit of size (default, arcsec, arcmin, degree, radian, physical, image)
        angle: for box, ellipse
        point: point type (only for point)
        coord: coordinate name for X, Y
        coord_unit: (default, degree, radian, physical, image)
        **kargs: other properties for ds9 region (color etc.)
        """
        if type == None:
            type=self.type
        else:
            self.type=type
        if size_x == None:
            size_x=self.size_x
        else:
            self.size_x=size_x
        if size_y == None:
            size_y=self.size_y
        else:
            self.size_y=size_y
        if size_unit == None:
            su=self.su
        else:
            su=getsu(size_unit)
            self.su=su
        if point == None:
            point=self.point
        else:
            self.point=point
        if coord == None:
            coord=self.coord
        else:
            self.coord=coord
        if coord_unit == None:
            cu=self.cu
        else:
            cu=getcu(coord_unit)
            self.cu=cu
        
        if len(self.list) > 0:
            tmp_list = []
            for sp in self.list:
                sp.type=type
                sp.size_x=size_x
                sp.point=point
                sp.coord=coord
                sp.su=su
                sp.cu=cu
                tmp_props = []
                for k, v in kwargs.iteritems():
                    if sp.regprops.has_key(k):
                        if v != None: tmp_props.append((k, kwargs[k]))
                        else: tmp_props.append((k, sp.regprops[k]))
                    else: tmp_props.append((k, v))
                sp.regprops=dict(tmp_props)
                tmp_list.append(sp)
            self.list = tmp_list
        else:
            pass
    
    def printreg(self):
        for sp in self.list:
            print sp.getreg()

class VectorList(ShapeList):
    """class VectorList
    list of vectors
    """
    def loaddata(self, datalist, coord='image', coord_unit='default',
                 xname=None, yname=None, lenname=None, anglename=None,
                 group='', **kwargs):
        self.coord=coord
        self.cu=getcu(coord_unit)
        self.groupid=group
        self.type='vector'
        index=0
        for ((x, y, xname, yname), len, ang) in zip(self._loadcoord(datalist, coord, xname, yname),
                                                  datalist.getarray(lenname), datalist.getarray(anglename)):
            idstring = "%s%04i"%(group, index)
            index=index+1
            obj = Shape(x, y, type='vector', length=len, angle=ang,
                        coord=coord, coord_unit=coord_unit, id=idstring, groupid=group, **kwargs)
            self.list.append(obj)

class LineList(ShapeList):
    """class LineList
    list of lines
    """
    def loaddata(self, datalist, coord='image', coord_unit='default',
                 xname=None, yname=None, x2name=None, y2name=None,
                 arrow1=False, arrow2=False,
                 group='', **kwargs):
        self.coord=coord
        self.cu=getcu(coord_unit)
        self.groupid=group
        self.type='line'
        index=0
        for ((x, y, xname, yname), x2, y2) in zip(self._loadcoord(datalist, coord, xname, yname),
                                                  datalist.getarray(x2name), datalist.getarray(y2name)):
            idstring = "%s%04i"%(group, index)
            index=index+1
            obj = Shape(x, y, type='line', x2=x2, y2=y2, arrow1=arrow1, arrow2=arrow2,
                        coord=coord, coord_unit=coord_unit, id=idstring, groupid=group, **kwargs)
            self.list.append(obj)

def reg2shape(region, coord):
    """Function reg2shape
    generate Shape object from region string
    """
    reg, opt = region.split('#')
    reg = reg.strip()
    opt = opt.strip()
    type, coords = reg.split('(')
    coordlist = coords[:-1].split(',')
    x, y = coordlist[:2]
    if(len(coordlist) > 2):
        size_x = coordlist[2]
    else:
        size_x = None
    if(len(coordlist) > 3):
        size_y = coordlist[3]
    else:
        size_y = None
    if(len(coordlist) > 4):
        angle = coordlist[4]
    else:
        angle= None
    
    kwargs = []
    tmplist = opt.split('=')
    key = tmplist[0]
    id = None
    groupid = None
    for optdata in tmplist[1:]:
        try:
            value, key1 = optdata.rsplit(' ', 1)
        except ValueError:
            value = optdata
        value = value.strip('"{}')
        if key == 'tag' and value.find(':')>0:
            tag_type, tag_value = value.split(':', 1)
            if tag_type == 'id':
                id = tag_value
            elif tag_type == 'gid':
                groupid = tag_value
            else:
                kwargs.append((key,value))
        else:
            kwargs.append((key,value))
        key = key1
    kwargs = dict(kwargs)
    
    if kwargs.has_key('point'):
        point = kwargs['point']
        del kwargs['point']
    else:
        point = None
    
    return Shape(x, y, coord=coord,
                 type=type, size_x=size_x, size_y=size_y, angle=angle,
                 point=point, id=id, **kwargs)
