# slit.py
# classes of slit
# 2006.06.30 YOSHIKAWA, Tomohiro
from mcsmdp.shape import *
from mcsmdp.error import *
from astrocat import mdp,datalist
from math import *

class Slit(Shape):
    def __init__(self, x, y, width_x, width_y, id, coord='image', angle=0, slit_type='slit',
                 name='', priority=1, objdata=[], **kwargs):
        """constructor
        x, y: slit position
        width_x, width_y: width of slit or hole
        angle: position angle of hole
        slit_type: slit or hole
        """
        self.slit_type = slit_type
        self.width_x = width_x
        self.width_y = width_y
        self.name = name
        self.priority = priority
        self.data = objdata
        #self.data = {'width_x':width_x,
        #        'width_y':width_y,
        #        'name':name}
        if slit_type == 'hole':
            size_x = width_x/2.0
            size_y = width_y/2.0
            type = 'ellipse'
        else:
            size_x = width_x
            size_y = width_y
            type = 'box'
        label = "{%s}"%(name)
        if coord == 'fk5':
            coord_unit = 'degree'
            size_unit = 'arcsec'
        else:
            coord_unit = 'image'
            size_unit = 'image'
        Shape.__init__(self, x, y, coord=coord, coord_unit=coord_unit,
                       type=type, size_x=size_x, size_y=size_y, size_unit=size_unit,
                       angle=angle, id=slit_type+':'+id, groupid=slit_type, text=label, **kwargs)
    def move(self, dx, dy):
        self.update(x=self.x + dx, y=self.y + dy)
    def mod(self, width_x, width_y, name=None):
        if name != None:
            self.name = name
        else:
            name = self.name
        size_x = None
        size_y = None
        if width_x != None:
            self.width_x = width_x
            
            if self.type == 'ellipse':
                size_x = width_x/2.0
            else:
                size_x = width_x
                self.type = 'box'
        if width_y != None:
            self.width_y = width_y
            if self.type == 'ellipse':
                size_y = width_y/2.0
            else:
                size_y = width_y
                self.type = 'box'
        self.update(size_x=size_x, size_y=size_y, text="{"+name+"}")
    def delete(self):
        self.enable=False
        #Shape.hide(self)
    def setspec(self, spec):
        self.spec=spec

    ## obsolete
    def show(self, move, rotate, edit, delete):
        if self.enable == True:
            Shape.show(self, move, rotate, edit, delete)

    def showspec(self):
        if self.enable: self.spec.show()

class SpecList(ShapeList):
    def loaddata(self, specdata, group='', **kwargs):
        index=0
        self.groupid = group
        self.list=[]
        for data in specdata:
            idstring="spec:%d"%(index)
            index=index+1
            spec=Shape(data['X'], data['Y'], type='box',
                       size_x=data['WIDTH_X'], size_y=data['WIDTH_Y'],
                       coord='image', coord_unit='', id=idstring, groupid=group, **kwargs)
            self.list.append(spec)

class SlitList(ShapeList):
    def __init__(self, inst):
        """SlitList
        inst: instrument object
        """
        self.inst=inst
        self.list=[]
        self.index=0
        ShapeList.__init__(self)
        self.setcolor('yellow', 'cyan')

    def setcolor(self, holecolor, speccolor):
        self.holecolor = holecolor
        self.speccolor = speccolor
        
    def loaddata(self, slitdata, coord='image', xname='X', yname='Y',
                 rotate=False, invert_y=False, invert_x=False, xlen=2048, ylen=2096,
                 holecolor='yellow', speccolor='cyan',
                 overwrite=True, **kwargs):
        """method loaddata
        datalist: MDP datalist
        rotate: rotate -90 degree
        invert_x: invert symmetrically with respect to the x-axis
        invert_y: invert symmetrically with respect to the y-axis
        xlen, ylen: x, y length of the image
        **kwargs: other properties for ds9 region

        sequence of conversion:
        1. invert_x
        2. invert_y
        3. rotate
        FOCAS to MOIRCS conversion: rotate=True, invert_x=True
        """
        if overwrite: self.list=[]
        self.slitargs = kwargs
        self.holecolor = holecolor
        self.speccolor = speccolor
        kwargs2 = []
        for k,v in kwargs.iteritems():
            if k == 'color': continue
            kwargs2.append((k,v))
        kwargs2 = dict(kwargs2)
        for (x, y, xname, yname), data in zip(self._loadcoord(slitdata, coord=coord, xname=xname, yname=yname), slitdata):
            if invert_x:
                x1 = xlen-x
                x = x1
            if invert_y:
                y1 = ylen-y
                y = y1
            if rotate:
                x1 = y
                y1 = x
                y2 = xlen - y1
                x = x1
                y = y2
                width_x = data['WIDTH_Y']
                width_y = data['WIDTH_X']
            else:
                width_x = data['WIDTH_X']
                width_y = data['WIDTH_Y']
            idstring = str(self.index)
            self.index = self.index+1
            objdata = []
            for d in data:
                k = d[0]
                if k!=xname and k!=yname and k!='WIDTH_X' and k!='WIDTH_Y' and k!='ANGLE' and k!='PRIORITY' and k!='SLIT_TYPE' and k!='NAME':
                    objdata.append(d)
            if data['SLIT_TYPE'] == 'hole':
                sl = Slit(x, y, width_x=width_x, width_y=width_y, id=idstring, coord=coord,
                          angle=data['ANGLE'], slit_type=data['SLIT_TYPE'],
                          name=data['NAME'], priority=data['PRIORITY'],
                          color=holecolor, objdata=objdata, **kwargs2)
            else:
                sl = Slit(x, y, width_x=width_x, width_y=width_y, id=idstring, coord=coord,
                          angle=data['ANGLE'], slit_type=data['SLIT_TYPE'],
                          name=data['NAME'], priority=data['PRIORITY'],
                          objdata=objdata, **kwargs)
            self.list.append(sl)
        self.setspec(color=speccolor)
    def setspec(self, color=None, **kwargs):
        """setspec
        set spectrum list from slit list
        kargs: ds9 options for spectrum
        """
        if color == None:
            color=self.speccolor
        smax, smin = self.inst.disperser.getdispersion()
        self.speclist = SpecList()
        slist = []
        if self.inst.dispaxis == 1:
            for sl in self.list:
                if sl.enable == False:continue
                if(sl.x > self.inst.center_x): #ch2
                    slist.append({'X': (2*sl.x+smax+smin)*0.5,
                                  'Y': sl.y,
                                  'WIDTH_X': smax-smin,
                                  'WIDTH_Y': sl.width_y
                                  })
                else: #ch1
                    slist.append({'X': (2*sl.x-smax-smin)*0.5,
                                   'Y': sl.y,
                                  'WIDTH_X': smax-smin,
                                  'WIDTH_Y': sl.width_y
                                  })
        else:
            for sl in self.list:
                if sl.enable == False:continue
                if(sl.y > self.inst.center_y): #ch2
                    slist.append({'Y': (2*sl.y+smax+smin)*0.5,
                                  'X': sl.x,
                                  'WIDTH_Y': smax-smin,
                                  'WIDTH_X': sl.width_x
                                  })
                else: #ch1
                    slist.append({'Y': (2*sl.y-smax-smin)*0.5,
                                  'X': sl.x,
                                  'WIDTH_Y': smax-smin,
                                  'WIDTH_X': sl.width_x
                                  })
        self.speclist.loaddata(slist, group='spec', color=color, **kwargs)
    def modbyid(self, id, width_x=None, width_y=None, name=None):
        for index in range(len(self.list)):
            if self.list[index].id == id:
                self.list[index].mod(width_x, width_y, name)
                self.setspec(color=self.speccolor)
                return
        raise MCSMDPError, "no such id: "+str(id)
    def movebyid(self, id, dx, dy):
        for index in range(len(self.list)):
            if self.list[index].id == id:
                self.list[index].move(dx, dy)
                self.setspec(color=self.speccolor)
                return
        raise MCSMDPError, "no such id: "+str(id)
    def delbyid(self, id):
        for index in range(len(self.list)):
            if self.list[index].id == id:
                self.list[index].delete()
                self.setspec(color=self.speccolor)
                return
        raise MCSMDPError, "no such id: "+str(id)
    def addslit(self, x, y, width_x, width_y, angle, slit_type,
                name, priority, **kwargs):
        idstring = str(self.index)
        self.index = self.index+1
        if slit_type == 'hole':
            kwargs2 = []
            for k,v in kwargs.iteritems():
                if k == 'color': continue
                kwargs2.append((k,v))
            kwargs2 = dict(kwargs2)
            sl = Slit(x, y, width_x=width_x, width_y=width_y,
                      angle=angle, slit_type=slit_type,
                      name=name, priority=priority,
                      id=idstring, color=self.holecolor, **kwargs)
        else:
            sl = Slit(x, y, width_x=width_x, width_y=width_y,
                      angle=angle, slit_type=slit_type,
                      name=name, priority=priority,
                      id=idstring, **kwargs)
        self.list.append(sl)
        self.setspec(color=self.speccolor)
    def showspec(self):
        self.speclist.show()
    def hidespec(self):
        self.speclist.hide()
    
    def xyshift(self, shift_x, shift_y):
        for index in range(len(self.list)):
            self.list[index].x = self.list[index].x + shift_x
            self.list[index].y = self.list[index].y + shift_y
    def magnify(self, mag_x, mag_y):
        for index in range(len(self.list)):
            self.list[index].x = self.list[index].x * mag_x
            self.list[index].y = self.list[index].y * mag_y
    def rotate(self, rotation):
        rotation = rotation/180.0*pi
        for index in range(len(self.list)):
            x0 = self.list[index].x
            y0 = self.list[index].y
            x1 = x0*cos(rotation) - y0*sin(rotation)
            y1 = x0*sin(rotation) + y0*cos(rotation)
            self.list[index].x = x1
            self.list[index].y = y1
    def slit_swapxy(self):
        for slit in self.list:
            ys_new = slit.size_x
            xs_new = slit.size_y
            slit.size_x = xs_new
            slit.size_y = ys_new
            yw_new = slit.width_x
            xw_new = slit.width_y
            slit.width_x = xw_new
            slit.width_y = yw_new
    
#    def setspec(self, inst, **kwargs):
#        """set spectrum
#        inst: an instance of Instrument class
#        kargs: ds9 options for spectrum
#        """
#        spec_max, spec_min = inst.disperser.getdispersion()
#        for index in range(len(self.list)):
#            sl = self.list[index]
#            if inst.dispaxis == 'X':
#                width_x = spec_max-spec_min
#                width_y = sl.width_y
#                x = x + (spec_max+spec_min)/2
#                y = y
#            else:
#                width_y = spec_max-spec_min
#                width_x = sl.width_x
#                y = y + (spec_max+spec_min)/2
#                x = x
#            sl.setspec(Shape(x, y, coord='image', coord_unit='image',
#                             type='box', size_x=width_x, size_y=width_y, size_unit='image',
#                             angle=0, id='spec:'+sl.id, **kwargs))
        
#    def showspec(self):
#        for index in range(len(self.list)):
#            self.list[index].showspec()
    def getdatalist(self):
        retlist=[]
        for index in range(len(self.list)):
            if self.list[index].enable:
                retlist.append({
                    'X':self.list[index].x,
                    'Y':self.list[index].y,
                    'WIDTH_X':self.list[index].width_x,
                    'WIDTH_Y':self.list[index].width_y,
                    'ANGLE':self.list[index].angle,
                    'PRIORITY':self.list[index].priority,
                    'SLIT_TYPE':self.list[index].slit_type,
                    'NAME':self.list[index].name})
        return datalist.DataList(retlist)
    def writemdp(self, file, xmin=None, xmax=None, ymin=None, ymax=None):
        """writemdp
        file: output file name
        """
        mdpdata = self.getdatalist()
        if xmin != None and xmax != None:
            mdpdata = mdpdata.thresh_list('X', xmax, xmin)
        if ymin != None and ymax != None:
            mdpdata = mdpdata.thresh_list('Y', ymax, ymin)
        mdp.writedata(mdpdata, file)
