"""sncslc - S/N calculator for spectroscopy
2008.02.28. YOSHIKAWA, Tomohiro
$Id: sncalc.py,v 1.1 2008-03-19 10:33:54 tomohiro Exp $
"""
import numpy
import pyfits
import specdata
from astrotool import cosmology

def sncalc(center, flux, fwhm, width, t, nsum,
           rn = 18, nexp=10,
           sky_fitsname='mask3_HK500/HK500_sBzK_MIPS_GT3ch1_33.fits',
           eta_fitsname='../ana_v00/STANDARD/StdMask3_HK500resa.fits',
           ):
    """
    center: list of center wavelength of emission to calculate
    flux: total flux of the emission
    fwhm: FWHM of the emission
    width: width of aperture to calculate the S/N (phi=fwhm*width)
    t: exposure time
    nsum: number of pixels to summing along a slit
    rn: readout noise (e-)
    nexp: number of exposure
    sky_fitsname: 1D FITS file name of sky background
    eta_fitsname: 1D FITS file name of transmission curve
    """
    ## background
    #sky_fits = pyfits.open(sky_fitsname)
    #exptime = sky_fits[0].header['EXPTIME']
    #gain= sky_fits[0].header['GAIN']
    #cdelt= sky_fits[0].header['CDELT1']
    #sky0 = twodspec2onedspec(sky_fits[1], 30, 100)
    #sky = specdata.spec_arith(sky0, '/', exptime*gain)

    sky0, shead = pyfits.getdata(sky_fitsname, header=True)
    exptime = shead['EXPTIME']
    gain= shead['GAIN']
    cdelt= shead['CDELT1']
    sky = specdata.spec_arith(sky0, '/', exptime*gain)
    
    
    eta_fits = pyfits.open(eta_fitsname)
    vlist = []
    for c in center:
        ## signal
        fwhm_w = fwhm / 3e5 * c
        wmin = c - fwhm_w*width/2
        wmax = c + fwhm_w*width/2
        #print wmin, wmax
        widx = (sky.field('WAVELENGTH') > wmin) & (sky.field('WAVELENGTH') < wmax)
        #return widx
        wl = sky[widx].field('WAVELENGTH')
        model0 = specdata.mkspecdata(wl, gaussian(wl, c, flux, fwhm))
        
        model = specdata.spec_arith(model0, '*', eta_fits[1].data)
        ## readout noise
        npix = (wmax-wmin)/cdelt
        readnoise = rn/gain*numpy.sqrt(2*nexp*npix*nsum)/t

        ## S/N
        s = specdata.integrate(wl, model.field('FLUX'))
        n = numpy.sqrt(specdata.integrate(wl, sky[widx].field('FLUX')*2/t*nsum))

        vlist.append([c,s,n,readnoise])

    #return vlist
    return numpy.rec.array(vlist, names=['wl', 'signal', 'bgnoise', 'rdnoise'])

    print 'read out noise: %f'%(readnoise)
    print 'background noise: %f'%(n)
    return s/n

def sncalc2(center, sfr, fwhm, width, t, nsum,
            rn = 18, nexp=10,
            sky_fitsname='mask3_HK500/HK500_sBzK_MIPS_GT3ch1_33.fits',
            eta_fitsname='../ana_v00/STANDARD/StdMask3_HK500resa.fits',
            ):
    ## background
    sky_fits = pyfits.open(sky_fitsname)
    exptime = sky_fits[0].header['EXPTIME']
    gain= sky_fits[0].header['GAIN']
    cdelt= sky_fits[0].header['CDELT1']
    sky0 = twodspec2onedspec(sky_fits[1], 30, 100)
    sky = specdata.spec_arith(sky0, '/', exptime*gain)

    ## cosmology
    cosmol = cosmology.Cosmology()
    
    vlist = []
    for c in center:
        ## signal
        fwhm_w = fwhm / 3e5 * c
        wmin = c - fwhm_w*width/2
        wmax = c + fwhm_w*width/2
        widx = (sky.field('WAVELENGTH') > wmin) & (sky.field('WAVELENGTH') < wmax)
        wl = sky[widx].field('WAVELENGTH')
        z = c/6563. - 1
        flux = sfr/7.9e-42/(4*numpy.pi*(cosmol.dl(z)*100)**2)
        model0 = specdata.mkspecdata(wl, gaussian(wl, c, flux, fwhm))
        eta_fits = pyfits.open(eta_fitsname)
        model = specdata.spec_arith(model0, '*', eta_fits[1].data)
        ## readout noise
        npix = (wmax-wmin)/cdelt
        readnoise = rn/gain*numpy.sqrt(2*nexp*npix*nsum)/t
        
        ## S/N
        s = specdata.integrate(wl, model.field('FLUX'))
        n = numpy.sqrt(specdata.integrate(wl, sky[widx].field('FLUX')*2/t*nsum))

        vlist.append([c,s,n,readnoise,z,flux])

    #return vlist
    return numpy.rec.array(vlist, names=['wl', 'signal', 'bgnoise', 'rdnoise', 'redshift', 'flux'])

##############
# SUBROUTINE #
##############

def gaussian(x, center, flux, fwhm):
    """
    Gaussian profile
    x: x array
    center: center wavelength
    flux: total flux
    fwhm: FWHM in km/s
    """
    fwhm_l = fwhm/2.998e5*center
    s = fwhm_l/(2.*numpy.sqrt(2.*numpy.log(2.)))
    return flux*numpy.exp(-(x-center)**2/2/s**2)/numpy.sqrt(2*numpy.pi)/s

def twodspec2onedspec_file(fname, outname, ymin=30, ymax=100):
    sky_fits = pyfits.open(fname)
    wl, fl = twodspec2onedspec(sky_fits[1], ymin, ymax)
    header = sky_fits[0].header
    header.update('SKY_ORG', fname)
    header.update('SKY_YMIN', ymin)
    header.update('SKY_YMAX', ymax)
    specdata.mkspecdata(wl, fl, outfile=outname, header=header)

def twodspec2onedspec(tfits, ymin, ymax):
    crpix = tfits.header['CRPIX1']
    crval = tfits.header['CRVAL1']
    cdelt = tfits.header['CDELT1']
    
    x = numpy.arange(tfits.header['NAXIS1']) + 1
    wl = cdelt * (x - crpix) + crval

    fl = 0
    num = 0
    for line in tfits.data[ymin:ymax]:
        fl += line
        num += 1
    fl = fl / num / cdelt
    return wl, fl
