"""
specdata - spectrum data tools
2007.06.19. YOSHIKAWA, Tomohiro
$Log: not supported by cvs2svn $
Revision 1.5  2008/02/01 23:54:20  tomohiro
move integrate() and interpolate to integrate.py

Revision 1.4  2007/08/24 07:38:24  tomohiro
load_onedspec:
- add pera option
- add header to output specdata file
spec_arith:
- support pyfits.NP_pyfits.FITS_rec for op2

Revision 1.3  2007/08/02 06:48:53  tomohiro
modify edge processing in interpolate()

Revision 1.2  2007/08/01 02:31:47  tomohiro
mkspecdata: add background, header option
writeto: add header option
spec_arith: add outfile option
twodspec_arith: modify to allow numeric for op2
delete obsolete functions

Revision 1.1  2007/07/17 07:21:19  tomohiro
spectrum data tools

"""
import pyfits
import numpy
import mcsmdp
import os
from integrate import integrate,interpolate
from numpy.core.records import recarray
from pyfits import FITS_rec

c=2.99792458e18 # speed of light (A/s)

def mkspecdata(wl, flux, noise=None, background=None, outfile=None, filter=False, header=None):
    if filter:
        vlname = 'RESPONSE'
        unit = 'fraction'
    else:
        vlname = 'FLUX'
        unit = 'erg s^-1 cm^-2 A^-1'
    c_wl = pyfits.Column(name='WAVELENGTH', format='D', unit='Angstrom', array=wl)
    c_vl = pyfits.Column(name=vlname, format='D', unit=unit, array=flux)
    outtable = [c_wl, c_vl]
    if noise != None:
        c_ns = pyfits.Column(name='NOISE', format='D', unit=unit, array=noise)
        outtable += [c_ns]
    if background != None:
        c_bg = pyfits.Column(name='BACKGROUND', format='D', unit=unit, array=background)
        outtable += [c_bg]

    out_f = pyfits.new_table(outtable, header=header)
    
    if outfile != None:
        out_f.writeto(outfile)
    else:
        return out_f.data

def writeto(spec, outfile, format='fits', header=None, label=None, integ=None):
    """
    write specdata to file
    spec: specdata
    outfile: output file name
    format: output format
    - fits: binary table FITS file
    - specfit: input file for specfit

    ## fits options ##
    - header: header table to write 
    
    ## specfit options ##
    - label: label for specfit format
    - integ: integtime for specfit format
    """
    if format == 'fits':
        pyfits.new_table(spec, header=header).writeto(outfile)
    elif format == 'specfit':
        if label == None: label = os.path.splitext(outfile)[0]
        outf = open(outfile, 'w')
        outf.write(label+'\n')
        if integ == None:
            outf.write('\t%d\n'%(len(spec)))
        else:
            outf.write('\t%d %d\n'%(len(spec), integ))
        for wl, fl, er in zip(spec.field('WAVELENGTH'),
                              spec.field('FLUX'),
                              spec.field('NOISE')):
            outf.write('%f\t%e\t%e\n'%(wl,fl,er))
        outf.close()

def load_onedspec(file, outfile=None, pera=False):
    """
    load specdata from onedspec fits file
    file: onedspec fits file
    outfile: fits table file name to save
    pera: calculate in ADU/A
    """
    fits_f = pyfits.open(file)
    crpix = fits_f[0].header['CRPIX1']
    crval = fits_f[0].header['CRVAL1']
    try:
        cdelt = fits_f[0].header['CDELT1']
    except KeyError:
        print "header keyword \'CDELT1\' not found. use CD1_1 instead"
        cdelt = fits_f[0].header['CD1_1']
    
    x = numpy.arange(fits_f[0].header['NAXIS1']) + 1
    wl = cdelt * (x - crpix) + crval
    if pera:
        fl = numpy.array(fits_f[0].data) / cdelt
    else:
        fl = numpy.array(fits_f[0].data)
    return mkspecdata(wl, fl, outfile=outfile, header=fits_f[0].header)

def spec_arith(op1, op, op2, outfile=None,
               key_op1_wl='WAVELENGTH',key_op1_fl='FLUX',key_op2_wl='WAVELENGTH',key_op2_fl='FLUX',
               filter=False):
    """
    calculate op1 op([+-*/]) op2
    op2 can be an scalar
    key_op[12]_wl: key name of wavelength for op[12]
    key_op[12]_fl: key name of flux for op[12]
    Note: op2 is interpolated to op1
    """
    if filter:
        key_op1_fl = 'RESPONSE'
        key_op2_fl = 'RESPONSE'
    op1_wl = op1.field(key_op1_wl)
    op1_fl = op1.field(key_op1_fl)
    if type(op2) == recarray or type(op2) == FITS_rec:
        op2_wl = op2.field(key_op2_wl)
        op2_fl = op2.field(key_op2_fl)
        op2_fli = interpolate(op2_fl, op2_wl, op1_wl)
    else:
        op2_fli = op2
    
    if op == '+':
        ans_fl = op1_fl + op2_fli
    if op == '-':
        ans_fl = op1_fl - op2_fli
    if op == '*':
        ans_fl = op1_fl * op2_fli
    if op == '/':
        ans_fl = op1_fl / op2_fli
    
    return mkspecdata(op1_wl, ans_fl, outfile=outfile, filter=filter)

def twodspec_arith(opfits, op, op2,
                   key_op2_wl='WAVELENGTH',key_op2_fl='FLUX',
                   ):
    """
    calculate opfile (2D spec FITS) op([+-*/]) op2
    op2 shold be specdata instance or int or float
    key_op2_wl: key name of wavelength for op2
    key_op2_fl: key name of flux for op2
    Note: op2 is interpolated to op1
    """
    if type(op2) == recarray or type(op2) == FITS_rec:
        crpix = opfits.header['CRPIX1']
        crval = opfits.header['CRVAL1']
        try:
            cdelt = opfits.header['CDELT1']
        except KeyError:
            print "header keyword \'CDELT1\' not found. use CD1_1 instead"
            cdelt = opfits.header['CD1_1']
        x = numpy.arange(opfits.header['NAXIS1']) + 1
        op1_wl = cdelt * (x - crpix) + crval
        
        op2_wl = op2.field(key_op2_wl)
        op2_fl = op2.field(key_op2_fl)
        op2_fli = interpolate(op2_fl, op2_wl, op1_wl)
    else:
        op2_fli = op2
    
    if op == '+':
        for op1_fl in opfits.data: op1_fl += op2_fli
    if op == '-':
        for op1_fl in opfits.data: op1_fl -= op2_fli
    if op == '*':
        for op1_fl in opfits.data: op1_fl *= op2_fli
    if op == '/':
        for op1_fl in opfits.data: op1_fl /= op2_fli
    
    return opfits

def bandpass(spec, filter, fnu=False,
             key_spec_wl='WAVELENGTH',key_spec_fl='FLUX',
             key_filter_wl='WAVELENGTH',key_filter_res='RESPONSE',
             photon=False,
             debug=False):
    """
    calculate bandpass
    photon: calculate photon based mean if True, calculate energy based one if False
    """
    spec_wl = spec.field(key_spec_wl)
    spec_fl = spec.field(key_spec_fl)
    filter_wl  = filter.field(key_filter_wl)
    filter_res = filter.field(key_filter_res)
    
    filter_resi = interpolate(filter_res, filter_wl, spec_wl, boundary='constant')
    #return spec_wl, filter_resi, spec_fl
    if fnu:
        if photon:
            sp = integrate(spec_wl, spec_fl*filter_resi*spec_wl)
            fi = integrate(spec_wl, c*filter_resi/spec_wl)
        else:
            sp = integrate(spec_wl, spec_fl*filter_resi)
            fi = integrate(spec_wl, c*filter_resi/spec_wl**2)
    else:
        if photon:
            sp = integrate(spec_wl, spec_fl*filter_resi*spec_wl)
            fi = integrate(spec_wl, filter_resi*spec_wl)
        else:
            sp = integrate(spec_wl, spec_fl*filter_resi)
            fi = integrate(spec_wl, filter_resi)
    if debug: print sp, fi
    return sp/fi

def addfnu(spec, add=False, outfile=None):
    c=2.99792458e18 # speed of light (A/s)
    ar_wl = spec.field('WAVELENGTH') # wavelength (Angstrom)
    ar_fl = spec.field('FLUX') # F_lambda (erg s^-1 cm^-2 A^-1)
    
    ar_fn = ar_fl * (ar_wl**2) / c
    ar_fr = c/ar_wl
    
    c_fn = pyfits.Column(name='Fnu', format='D', unit='erg s^-1 cm^-2 Hz^-1', array=ar_fn)
    c_fr = pyfits.Column(name='FREQUENCY', format='D', unit='Hz', array=ar_fr)
    if add:
        c_wl = pyfits.Column(name='WAVELENGTH', format='D', unit='Angstrom', array=ar_wl)
        c_fl = pyfits.Column(name='FLUX', format='D', unit='erg s^-1 cm^-2 A^-1', array=ar_fl)
        out_f = pyfits.new_table([c_wl, c_fl, c_fr, c_fn])
    else:
        out_f = pyfits.new_table([c_fr, c_fn])
    if outfile != None:
        out_f.writeto(outfile)
    return out_f.data
