""" specstack.py
2007.02.01 YOSHIKAWA, Tomohiro
spectrum stacking analysis
$Log: not supported by cvs2svn $
Revision 1.6  2008/09/22 23:59:12  tomohiro
add luminosity and weight option to bootstrap()

Revision 1.5  2008/08/30 04:28:56  tomohiro
modify to retry mdpextract in bootstrap()

Revision 1.4  2008/06/20 04:32:38  tomohiro
modified bootstrap() for [OIII] fitting

Revision 1.3  2008/06/14 02:13:23  tomohiro
add bootstrap() for Halpha emission

"""
import os, sys
from pyraf import iraf
import pyfits, numpy
from rpy import r
from numpy import rec

from mcsmdp import error
from astrotool import cosmology
from iraftool import fileutil, sfdb
import integrate, specdata

iraf.mcsmdp()

def specstack(flist, zlist=None, klist=None ,wl=None,
              luminosity=True, weight=True, interpol_method='linear'):
    if klist == None:
        klist = numpy.zeros(len(flist))
    if zlist == None:
        zlist = numpy.zeros(len(flist))
    
    # cosmology
    csm = cosmology.Cosmology()
    
    # get base wavelangth
    if wl == None:
        data = pyfits.getdata(flist[0], ext=1)
        w0 = data.field('WAVELENGTH') / (1+zlist[0])
    else:
        w0 = numpy.arange(*wl)
    print 'wl max: %s'%(w0.max())
    print 'wl min: %s'%(w0.min())
    
    ysum = numpy.zeros(len(w0)) # sum (K*W*L)
    ssum = numpy.zeros(len(w0)) # sum (K*W*S)**2
    wsum = numpy.zeros(len(w0)) # sum (W)
    dl0 = csm.dl(zlist[0])*100
    weffsum = numpy.array([True]*len(w0))
    
    for f, z, k in zip(flist, zlist, klist):
        # luminosity distance
        dl = csm.dl(z)*100
        
        data = pyfits.getdata(f, ext=1)
        # interpolate and calculate flux
        wr = data.field('WAVELENGTH')/(1+z)
        if luminosity:
            yy = integrate.interpolate(data.field('FLUX'), wr, w0,
                                       boundary='constant', constant=0, method=interpol_method) * (1+z) * 4 * numpy.pi * dl**2
            sigma_yy  = integrate.interpolate(data.field('NOISE'), wr, w0,
                                              boundary='nan', method=interpol_method) * (1+z) * 4 * numpy.pi * dl**2
        else:
            yy = integrate.interpolate(data.field('FLUX'), wr, w0,
                                       boundary='constant', constant=0, method=interpol_method)
            sigma_yy  = integrate.interpolate(data.field('NOISE'), wr, w0,
                                              boundary='nan', method=interpol_method)

        # effective wavelength
        weff = (sigma_yy > 0) & (sigma_yy != numpy.nan) & (w0 > wr.min()) & (w0 < wr.max())
        sigma_yy1 = numpy.zeros(len(w0))
        sigma_yy1[weff] = sigma_yy[weff]
        
        # weight
        if weight:
            wg = numpy.zeros(len(w0))
            wg[weff] = 1./sigma_yy[weff]**2
        else:
            wg = numpy.zeros(len(w0))
            wg[weff] = 1

        #scale
        if k != 0:
            if luminosity:
                scale = klist[0]/k * dl0**2 / dl**2
            else:
                scale = klist[0]/k
            print scale
        else:
            scale = 1
        
        # stack
        ysum += scale * wg * yy
        ssum += (scale * wg * sigma_yy1)**2
        wsum += wg
        #print len(weffsum), len(weff)
        weffsum = weffsum|weff

    #return w0, ysum, ssum, wsum
    #return w0, ysum/wsum, numpy.sqrt(ssum/(wsum**2))
    return w0[weffsum], ysum[weffsum]/wsum[weffsum], numpy.sqrt(ssum[weffsum]/(wsum[weffsum]**2))

c = 2.99792458e5 # speed of light (km/s)
retry_max = 3    # number of retry for mdpextract (this should be more than 2)

def bootstrap(infiles, zlist, klist=None, wl=(6450, 6650, 1.), luminosity=False, weight=True,
              paralist=None, parakey=None, 
              fittype='Halpha', name='INDEF', verbose=True, fit_interactive=True,
              boot_interactive=False, type_of_fit='simplex',
              iflux=iraf.INDEF, icenter=iraf.INDEF, ifwhm=iraf.INDEF,
              center_tolerance=100, fix_fwhm=False, cont_flux=0, cont_kmax=4.0e-22,
              sub_feature=True, sub_frac=iraf.INDEF, sub_fwhm=iraf.INDEF,
              o3_4959=True, num_boot=1000, ranseed=None,
              interpol_method='linear'
              ):
    
    specfile = fileutil.osimg(iraf.mktemp("tmp$"+'bsspecfile'))
    logfile  = iraf.osfn(iraf.mktemp("tmp$"+'bslogfile'))
    print 'specfile: ',specfile
    print 'logfile: ', logfile
    
    guess_OK = False
    while(not guess_OK):
        # estimate first guess
        if(os.access(specfile, os.F_OK)): os.unlink(specfile)
        if(os.access(logfile, os.F_OK)): os.unlink(logfile)
        
        wv, lm, sig = specstack(infiles, zlist, klist=klist, wl=wl, luminosity=luminosity, weight=weight, interpol_method=interpol_method)
        specdata.mkspecdata(wv, lm, noise=sig, outfile=specfile)
        
        retry = retry_max
        while(retry>0):
            try:
                iraf.mdpextract(infile=specfile, line=iraf.INDEF, center=iraf.INDEF, outfile='INDEF',
                                fittype=fittype, name=name,
                                fit_interactive=fit_interactive, type_of_fit=type_of_fit,
                                iflux=iflux, icenter=icenter, ifwhm=ifwhm,
                                center_tolerance=center_tolerance, fix_fwhm=fix_fwhm,
                                cont_flux=cont_flux, cont_kmax=cont_kmax,
                                sub_feature=sub_feature, sub_frac=sub_frac, sub_fwhm=sub_fwhm, logfile=logfile,
                                o3_4959=True)
                break
            except iraf.IrafError:
                retry -= 1
        if retry <= 0:
            raise error.MCSMDPError, 'ERROR: exceed # of retry for mdpextract:', retry
        
        if fit_interactive:
            print "Use these parameters for the first guess? [Yes/No/Cancel]:",
            ans = sys.stdin.readline()
            ans = ans.strip()
        else:
            ans = 'y'
        if len(ans) > 0:
            if ans[0] == 'Y' or ans[0] == 'y':
                guess_OK = True
            elif ans[0] == 'C' or ans[0] == 'c':
                if(os.access(specfile, os.F_OK)): os.unlink(specfile)
                if(os.access(logfile, os.F_OK)): os.unlink(logfile)
                return
    
    # get results and set to initial guess
    if fittype == 'Halpha':
        sftab = sfdb.sflogtab(logfile, ['test'], ['N/A'], linetype=sfdb.hatype)[0]
        
        iflux1 = sftab['Halpha-6563_flux']
        icenter1 = sftab['Halpha-6563_centroid']
        ifwhm1 = sftab['Halpha-6563_fwhm']/c*sftab['Halpha-6563_centroid']
        cont_flux1 = sftab['continuum_flux'] + sftab['continuum_slope']*(icenter - 1000)
    elif fittype == 'OIII':
        sftab = sfdb.sflogtab(logfile, ['test'], ['N/A'], linetype=sfdb.o3type)[0]
        
        iflux1 = sftab['[OIII]-5007_flux']
        icenter1 = sftab['[OIII]-5007_centroid']
        ifwhm1 = sftab['[OIII]-5007_fwhm']/c*sftab['[OIII]-5007_centroid']
        cont_flux1 = sftab['continuum_flux'] + sftab['continuum_slope']*(icenter - 1000)
    else:
        raise error.MCSMDPError, 'invalid fittype: '+fittype

    if ranseed != None:
        r.set_seed(ranseed)
    
    outlist = []
    if verbose:
        sout = 0
    else:
        sout = 1
    failcount = 0
    for b in range(num_boot):
        bidx = r.sample(range(len(infiles)), len(infiles), replace=True)
        flistb = infiles[bidx]
        zlistb = zlist[bidx]
        if klist is not None:
            klistb = klist[bidx]
        else:
            klistb = None
        #print flistb
        
        if(os.access(specfile, os.F_OK)): os.unlink(specfile)
        if(os.access(logfile, os.F_OK)): os.unlink(logfile)
        
        wv, lm, sig = specstack(infiles, zlistb, klist=klistb, wl=wl, luminosity=luminosity, weight=weight)
        specdata.mkspecdata(wv, lm, noise=sig, outfile=specfile)

        retry = retry_max
        while(retry>0):
            try:
                iraf.mdpextract(infile=specfile, line=iraf.INDEF, center=iraf.INDEF, outfile='INDEF',
                                fittype=fittype, name=name,
                                fit_interactive=boot_interactive, type_of_fit=type_of_fit,
                                iflux=iflux1, icenter=icenter1, ifwhm=ifwhm1,
                                center_tolerance=center_tolerance, fix_fwhm=fix_fwhm,
                                cont_flux=cont_flux1, cont_kmax=cont_kmax,
                                sub_feature=sub_feature, sub_frac=sub_frac, sub_fwhm=sub_fwhm, logfile=logfile,
                                o3_4959=True, Stdout=sout)
                break
            except iraf.IrafError:
                retry -= 1
        #print retry
        if retry <= 0:
            print 'exceed # of retry for mdpextract:', retry
            failcount += 1
            sflist = [None]*len(outlist[-1])
            #raise error.MCSMDPError, 'exceed # of retry for mdpextract'
        else:
            if fittype == 'Halpha':
                sftab = sfdb.sflogtab(logfile, ['test'], ['N/A'], linetype=sfdb.hatype)[0]
            elif fittype == 'OIII':
                sftab = sfdb.sflogtab(logfile, ['test'], ['N/A'], linetype=sfdb.o3type)[0]
            sflist = list(sftab.tolist())
        
        if paralist != None and parakey != None:
            mlist = []
            for k in parakey:
                mlist.append(r.median(paralist[bidx][k]))
            outlist.append(sflist+mlist)
        else:
            outlist.append(sflist)

    if(os.access(specfile, os.F_OK)): os.unlink(specfile)
    if(os.access(logfile, os.F_OK)): os.unlink(logfile)
    
    if paralist != None and parakey != None:
        outkeys = list(sftab.dtype.names) + parakey
    else:
        outkeys = sftab.dtype.names

    if failcount > 0:
        print "WARNING: failed specfit %d time(s)"%(failcount)
    return rec.array(outlist, names=outkeys)
